# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Freeze instance attributes into a read-only mode for thread saftey.
"""

from constants import SET_ERROR, DEL_ERROR


class Freezable(object):
    """
    Base class for classes that can be frozen into a read-only state.

    Be forewarned that Python does not offer true read-only attributes for
    user-defined classes.  Do *not* rely upon the read-only-ness of this
    class for security purposes!

    The point of this class is not to make it impossible to set or to delete
    attributes after an instance is frozen, but to make it impossible to do so
    *accidentally*.  Rather than constantly reminding ourselves of things like,
    "Don't set any attributes on this ``FooBar`` instance because doing so wont
    be thread-safe", this class offers a real way to enforce read-only attribute
    usage.

    For example, before a `Freezable` instance is frozen, you can set and delete
    its attributes as normal:

    >>> class Person(Freezable):
    ...     pass
    ...
    >>> p = Person()
    >>> p.name = 'John Doe'
    >>> p.phone = '123-456-7890'
    >>> del p.phone

    But after an instance is frozen, you cannot set its attributes:

    >>> isfrozen(p)  # Is this instance frozen?
    False
    >>> p.__freeze__()  # This will freeze the instance
    >>> isfrozen(p)
    True
    >>> p.department = 'Engineering'
    Traceback (most recent call last):
      ...
    AttributeError: frozen: cannot set Person.department to 'Engineering'

    Nor can you deleted its attributes:

    >>> del p.name
    Traceback (most recent call last):
      ...
    AttributeError: frozen: cannot delete Person.name

    However, as noted at the start, there are still obscure ways in which
    attributes can be set or deleted on a frozen `Freezable` instance.  For
    example:

    >>> object.__setattr__(p, 'department', 'Engineering')
    >>> p.department
    'Engineering'
    >>> object.__delattr__(p, 'name')
    >>> hasattr(p, 'name')
    False

    But again, the point is that you would never employ the above techniques
    *accidentally*.

    Lastly, this example aside, you should use the `freeze()` function rather
    than the `Freezable.__freeze__()` method.  For example:

    >>> f = Freezable()
    >>> isfrozen(f)
    False
    >>> freeze(f) is f  # freeze() returns the instance
    True
    >>> isfrozen(f)
    True
    """

    __frozen__ = False

    def __freeze__(self):
        if self.__frozen__:
            raise StandardError('already frozen: %r' % self)
        self.__frozen__ = True

    def __setattr__(self, name, value):
        """
        If unfrozen, set attribute named ``name`` to ``value``.

        If this instance is frozen, an AttributeError will be raised.

        :param name: Name of attribute to set.
        :param value: Value to assign to attribute.
        """
        if self.__frozen__:
            raise AttributeError(
                SET_ERROR % (self.__class__.__name__, name, value)
            )
        return object.__setattr__(self, name, value)

    def __delattr__(self, name):
        """
        If unfrozen, delete attribute named ``name``.

        If this instance is frozen, an AttributeError will be raised.

        :param name: Name of attribute to delete.
        """
        if self.__frozen__:
            raise AttributeError(
                DEL_ERROR % (self.__class__.__name__, name)
            )
        return object.__delattr__(self, name)


def freeze(instance, check=True):
    """
    Freeze an instance of the `Freezable` class or similar.

    This function can be used to freeze an instance of any class that implements
    the same freeze-protocol as the `Freezable` class.

    So that this function can be easily used within an assignment, ``instance``
    is returned after it's frozen.  For example:

    >>> f = Freezable()
    >>> f is freeze(f)
    True
    >>> f.attr = 'This wont work'
    Traceback (most recent call last):
      ...
    AttributeError: frozen: cannot set Freezable.attr to 'This wont work'

    Also see the `isfrozen()` function.

    :param instance: The instance of `Freezable` (or similar) to freeze.
    """
    if getattr(instance, '__frozen__', False):
        if check:
            raise StandardError('already frozen: %r' % (instance,))
        else:
            return instance
    if callable(getattr(instance, '__freeze__', None)):
        instance.__freeze__()
        assert instance.__frozen__ is True
    else:
        instance.__frozen__ = True
    return instance


def isfrozen(instance):
    """
    Return ``True`` if ``instance`` is frozen.

    For example:

    >>> f = Freezable()
    >>> isfrozen(f)
    False
    >>> f.__frozen__ = True
    >>> isfrozen(f)
    True

    Also see the `freeze()` function.

    :param instance: The instance of `Freezable` (or similar) to interrogate.
    """
    return getattr(instance, '__frozen__', False)
