# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
A pyments-based syntax highlighter and related widgets.
"""

from wehjit import builtins
pygments = None
if not getattr(builtins, '_skip_pygments', None):
    try:
        import pygments
    except ImportError:
        pass

if pygments:
    from textwrap import dedent
    from cStringIO import StringIO
    from genshi import XML
    from pygments import highlight
    from pygments.lexers import get_lexer_by_name, get_all_lexers
    from pygments.formatters import HtmlFormatter
    from pygments.styles import get_all_styles
    from wehjit import base
    from wehjit.util import Alternator
    from wehjit.plugins.page import QuickJump
    from wehjit.state import *


    class LexerInfo(object):
        def __init__(self, title, aliases, patterns, mimetypes):
            self.title = title
            self.aliases = tuple(aliases)
            self.name = self.aliases[0]
            self.patterns = tuple(patterns)
            self.mimetypes = tuple(mimetypes)

        def __repr__(self):
            return '%s(%r)' % (self.__class__.__name__, self.title)

    pygments_lexers = tuple(
        sorted(
            (LexerInfo(*args) for args in get_all_lexers()),
            key=lambda l: l.title
        )
    )

    pygments_lexer_aliases = set()
    for l in pygments_lexers:
        pygments_lexer_aliases.update(l.aliases)
    pygments_lexer_aliases = tuple(sorted(pygments_lexer_aliases))

    pygments_lexer_options = tuple(
        (l.name, l.title) for l in pygments_lexers
    )

    pygments_themes = tuple(sorted(get_all_styles()))

    pygments_theme_options = tuple((t, t) for t in pygments_themes)



    class Highlighter(base.Widget):
        """
        A pygments-based syntax highlighter.
        """

        code = Dynamic('code')

        def _validate_lexer(self, value):
            if value not in pygments_lexer_aliases:
                return 'text'
            return value
        lexer = Dynamic('lexer', validator=_validate_lexer, default='text')

        def _validate_theme(self, value):
            if value not in pygments_themes:
                return 'default'
            return value
        theme = Dynamic('theme', validator=_validate_theme, default='default')

        @DynamicProp
        def code_stream(self):
            if not self.__frozen__ or self.__request__:
                return self._highlight()
            return self.__code_stream

        def _highlight(self):
            if not self.code:
                return
            return XML(
                highlight(
                    self.code,
                    get_lexer_by_name(self.lexer),
                    HtmlFormatter(cssclass='%s %s' % (self.klass, self.theme))
                )
            )

        def __init__(self, **kw):
            super(Highlighter, self).__init__(**kw)
            self.__code_stream = None

        def _finalize_final(self):
            self.__code_stream = self._highlight()

        xml = """
        <div
            xmlns:py="http://genshi.edgewall.org/"
            py:if="code_stream"
            py:replace="code_stream"
        />
        """

        require_css = ('pygments.css',)

        @classmethod
        def register_assets(cls, assets):
            buf = StringIO()
            for style in pygments_themes:
                prefix = '%s.%s.%s ' % (cls.__tag__, cls.__name__, style)
                lines = HtmlFormatter(style=style).get_style_defs().splitlines()
                for line in lines:
                    buf.write(prefix + line + '\n')
            buf.reset()
            assets.register(fileobj=buf, key='pygments.css')

    builtins.register(Highlighter)


    misc_table_style = (
        ('', (
            ('font-size', '%(font_size_mono)s'),
            ('font-family', 'monospace'),
        )),
        ('pre', (
            ('margin', '0'),
        )),
        ('th', (
            ('vertical-align', 'top'),
            ('padding', '0.5em 1em'),
            ('text-align', 'right'),
        )),
        ('td', (
            ('vertical-align', 'top'),
            ('padding', '0.5em 1em'),
        )),
        ('tr', (
            ('background-color', '#eee'),
        )),
        ('tr.even', (
            ('background-color', '#ddd'),
        )),
        ('h2', (
            ('font-family', 'monospace'),
            ('font-weight', 'normal'),
            ('margin-top', '1.5em'),
            ('margin-bottom', '0'),
        )),
        ('h2 a', (
            ('text-decoration', 'none'),
            ('color', 'inherit'),
        )),
        ('h2 a:hover', (
            ('background-color', '#eee'),
        )),
        ('h2:target', (
            ('color', '#e02'),
        )),
    )


    class HighlighterLexers(base.Widget):
        """
        A list of pygments lexers.
        """

        @StaticProp
        def part_names(self):
            return ('aliases', 'patterns', 'mimetypes')

        @StaticProp
        def lexers(self):
            return pygments_lexers

        @DynamicProp
        def row(self):
            return Alternator(['even', None])

        xml = """
        <div
            xmlns:py="http://genshi.edgewall.org/"
            class="${css_classes}"
            id="${id}"
        >
        <div py:for="l in lexers">
        <h2 id="${l.name}"><a href="#${l.name}" py:content="l.title" /></h2>
        <table class="${row.reset()}">
        <tr py:for="name in part_names" class="${row.next()}">
        <th py:content="name" />
        <td py:content="repr(getattr(l, name))" />
        </tr>
        </table>
        </div>
        </div>
        """

        style = misc_table_style

    builtins.register(HighlighterLexers)


    class HighlighterLexerQuickJump(QuickJump):
        """
        Jump to anchor of a specific pygments lexer.
        """

        # StaticProp:
        _options = pygments_lexer_options

    builtins.register(HighlighterLexerQuickJump)


    class HighlighterThemeChanger(base.Widget):
        """
        Change pygments theme client-side.
        """

        label = Dynamic('label', default='Theme:')

        @StaticProp
        def options(self):
            return pygments_theme_options

        xml = """
        <label
            xmlns:py="http://genshi.edgewall.org/"
            class="${css_classes}"
            id="${id}"
        >
        ${label}
        <select name="${name}">
        <option
            py:for="(optvalue, optlabel) in options"
            value="${optvalue}"
            selected="${[None, 1][optvalue == 'default']}"
            py:content="optlabel"
        />
        </select>
        </label>
        """

        js_class = 'PygmentsThemeChanger'
        javascript = """
        Wehjit.bases.PygmentsThemeChanger = new Class({
            Extends: Wehjit.bases.FormField,

            on_change: function() {
                console.log('it worked', this.field.value);
                if (! this.highlighters ) {
                    console.log('Creating element list...');
                    this.highlighters = $$('div.Highlighter');
                    console.log('%d highlighters', this.highlighters.length);
                }
                this.highlighters.each(this.change_theme, this);

            },

            change_theme: function(el) {
                el.setProperty('class', 'Highlighter ' + this.field.value);
            },

        }
        );
        """

    builtins.register(HighlighterThemeChanger)


    class PluginBrowser(base.Container):
        """
        Display a set of `wehjit` plugins.
        """

        def __init__(self, **kw):
            super(PluginBrowser, self).__init__(**kw)
            for p in self.plugins:
                self.add(
                    self.new('PluginDetails', plugin=p, name=p.__name__)
                )

        plugins = Static('plugins', default=tuple())

        xml = """
        <div
            xmlns:py="http://genshi.edgewall.org/"
            class="${css_classes}"
            id="${id}"
        >
        <p py:content="'%d plugins registered' % len(plugins)" />
        <div py:for="child in children" py:replace="child.generate()" />
        </div>
        """

    builtins.register(PluginBrowser)


    def rstate(t):
        if t is None:
            return
        if not isinstance(t, tuple):
            return repr(t)
        if len(t) == 0:
            return 'tuple()'
        s = '(\n'
        for (name, default) in t:
            s += '    (%r, %r),\n' % (name, default)
        s += ')'
        return s

    def tup(t):
        if t is None:
            return
        if not isinstance(t, tuple):
            return repr(t)
        if len(t) == 0:
            return 'tuple()'
        s = '(\n'
        for i in t:
            s += '    %r,\n' % (i,)
        s += ')'
        return s

    def doc_dedent(p):
        if not p.__doc__:
            return
        tp = '"""'
        return ''.join([tp, '\n', dedent(p.__doc__.strip()), '\n', tp])


    class PluginDetails(base.Container):
        """
        Display the details of a particular `wehjit` plugin.
        """

        __pairs = (
            ('docstring', doc_dedent, 'python'),
            ('module', lambda p: p.__module__, 'python'),
            ('Static', lambda p: repr(p.Static), 'python'),
            ('Dynamic', lambda p: repr(p.Dynamic), 'python'),
            ('StaticProp', lambda p: repr(p.StaticProp), 'python'),
            ('DynamicProp', lambda p: repr(p.DynamicProp), 'python'),
            ('xml', lambda p: p.xml, 'html+genshi'),
        )

        plugin = Static('plugin')

        @DynamicProp
        def row(self):
            return Alternator(['even', None])

        xml = """
        <div
            xmlns:py="http://genshi.edgewall.org/"
            class="${css_classes}"
            id="${id}"
        >
        <h2 id="${plugin.__name__}">
        <a href="#${plugin.__name__}" py:content="plugin.__name__" />
        </h2>
        <table>
        <tr py:for="child in children" class="${row.next()}">
        <th py:content="child.name" />
        <td py:content="child.generate()" />
        </tr>
        </table>
        </div>
        """

        style = misc_table_style

        def __init__(self, **kw):
            super(PluginDetails, self).__init__(**kw)
            for (name, func, lexer) in self.__pairs:
                value = func(self.plugin)
                if value is None:
                    lexer = 'python'
                    value = repr(None)
                self.add(
                    self.new('Highlighter', name=name, code=value, lexer=lexer)
                )

    builtins.register(PluginDetails)
