# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Mange name-spaced CSS.
"""

import sys
from compat import json
from errors import StyleVarError


class SubDict(dict):
    def __init__(self, depth=0):
        self.__depth = depth

    def __repr__(self):
        return '%s(depth=%r)' % (self.__class__.__name__, self.__depth)

    def __getitem__(self, key):
        if key not in self:
            if self.__depth > 0:
                value = self.__class__(self.__depth - 1)
            else:
                value = dict()
            self[key] = value
        return super(SubDict, self).__getitem__(key)


def dict_diff(a, b, depth=0):
    if not a:
        return
    if not b:
        for item in a.iteritems():
            yield item
    else:
        for (key, value) in a.iteritems():
            if key not in b or value != b[key]:
                if depth > 0:
                    value = dict(
                        dict_diff(value, b.get(key), depth - 1)
                    )
                yield (key, value)


def filter_by_value(d, depth=0):
    if not d:
        return
    for key in sorted(d):
        value = d[key]
        if depth > 0:
            value = tuple(filter_by_value(value, depth - 1))
        if value:
            yield (key, value)


class Styles(object):
    """
    Manages tightly name-spaced CSS.
    """

    def __init__(self):
        self.reset()

    def reset(self):
        """
        Clear all style information.
        """
        self.__tree = SubDict(depth=1)
        self.__vars = dict()

    def __repr__(self):
        return '%s(<%d namespaces, %d vars>)' % (
            self.__class__.__name__, len(self.__tree), len(self.__vars)
        )

    def update(self, namespace, selector, definition):
        """
        Update style definition for *selector* in *namespace*.

        >>> s = Styles()
        >>> s.update('Menu', '> ul', {'display': 'none', 'position': 'fixed'})
        >>> s.update('Menu', ':hover > ul', {'display': 'block'})

        See `Styles.get()` and `Styles.diff()` for more information.

        :param namespace: The widget class name or instance id, e.g. ``'Menu'``
            or ``'screen1.menuset.menu1'``
        :param selector: The CSS sub-selector, e.g. ``''`` or ``':hover > ul'``
        :param definition: A ``dict`` mapping CSS property names to values
        """
        namespace = namespace.strip()
        selector = selector.strip()
        if isinstance(definition, dict):
            items = definition.iteritems()
        else:
            items = definition
        definition = dict(
            (k.strip(), v.strip()) for (k, v) in items
        )
        self.__tree[namespace][selector].update(definition)

    def update_global(self, selector, definition):
        """
        Update style for *selector* with *definition* in global namespace.

        This method is equivalent to calling `Styles.update()` with a
        *namespace* equal to ``''``.

        :param selector: The CSS sub-selector, e.g. ``''`` or ``':hover > ul'``
        :param definition: A ``dict`` mapping CSS property names to values
        """
        self.update('', selector, definition)

    def update_vars(self, **kw):
        """
        Update style-variables.

        This will override previously defined variables.  For example:

        >>> s = Styles()
        >>> s.update_vars(
        ...     color='#000',
        ...     background='#fff',
        ... )
        >>> s.get_vars()
        {'color': '#000', 'background': '#fff'}
        >>> s.update_vars(
        ...     color='#117',
        ...     font_size='12pt',
        ... )
        >>> s.get_vars()
        {'color': '#117', 'font_size': '12pt', 'background': '#fff'}

        :param kw: Keyword arguments providing values for style variables
        """
        self.__vars.update(kw)

    def get(self, namespace, substitute=True):
        """
        Retrieve all style definitions for ``namespace``.

        By default, style variables will be automatically substituted.  For
        example:

        >>> s = Styles()
        >>> s.update('Menu', '> a', {
        ...     'line-height': '2em',
        ...     'color': '%(menu_color)s',
        ... })
        >>> s.update_vars(menu_color='#fff')
        >>> s.get('Menu')
        {'> a': {'color': '#fff', 'line-height': '2em'}}

        But if called with ``substitute=False``, you get the raw CSS property
        values.  For example:

        >>> s.get('Menu', substitute=False)
        {'> a': {'color': '%(menu_color)s', 'line-height': '2em'}}

        `StyleVarError` will be raised if a CSS property value uses an undefined
        style variable.  For example:

        >>> s.update('Menu', '> a', {'background': '%(menu_background)s'})
        >>> s.get('Menu')
        Traceback (most recent call last):
          ...
        StyleVarError: undefined style var 'menu_background'
            used in CSS prop 'background' = '%(menu_background)s'
            in namespace 'Menu', selector '> a'

        :param namespace: The widget class name or instance id, e.g. ``'Menu'``
            or ``'screen1.menuset.menu1'``
        :param substitute: If ``True``, perform variable substitution
        """
        if namespace not in self.__tree:
            return
        ns = self.__tree[namespace]
        if substitute:
            return dict(
                (s, dict(self._substitute(namespace, s))) for s in ns
            )
        return dict(
            (s, dict(d)) for (s, d) in ns.iteritems()
        )

    def get_global(self, substitute=True):
        """
        Retrieve all style definitions for global namespace.
        """
        return self.get('', substitute)

    def get_vars(self):
        """
        Retrieve all style-variable.

        For example:

        >>> s = Styles()
        >>> s.get_vars()
        {}
        >>> s.update_vars(city='Boston', state='Massachusetts')
        >>> s.get_vars()
        {'city': 'Boston', 'state': 'Massachusetts'}
        """
        return dict(self.__vars)

    def dumps(self, indent=4):
        """
        Return all style information as a JSON-encoded string.

        For example:

        >>> s = Styles()
        >>> s.update('Menu', '', {'color': '%(color_menu)s'})
        >>> s.update_vars(color_menu='#600')
        >>> print s.dumps(indent=None)
        {"styles": {"Menu": {"": {"color": "%(color_menu)s"}}}, "vars": {"color_menu": "#600"}}

        :param indent: The indentation value to pass to ``json.dumps()``
        """
        return json.dumps(dict(styles=self.__tree, vars=self.__vars),
            sort_keys=True, indent=indent
        )

    def loads(self, data, reset=False):
        """
        Update style definitions from a JSON-encoded string *data*.

        For example:

        >>> s = Styles()
        >>> d = '{"styles": {"Menu": {"": {"color": "%(color_menu)s"}}}, "vars": {"color_menu": "#600"}}'
        >>> s.loads(d)
        >>> s.get('Menu', substitute=False)
        {u'': {u'color': u'%(color_menu)s'}}
        >>> s.get_vars()
        {u'color_menu': u'#600'}

        Normally the style information provided in *data* will be merged with
        the what is already present in the `Styles` instance.  For example:

        >>> s.loads('{"styles": {"Menu": {"": {"border": "1px solid black"}}}}')
        >>> s.get('Menu')
        {u'': {u'color': u'#600', u'border': u'1px solid black'}}

        However, if you call `Styles.loads()` with ``reset=True``,
        `Styles.reset()` will be called prior to loading the style information,
        resulting in the `Styles` instance only containing the style information
        defined in *data*.  For example:

        >>> s.loads('{"styles": {"Menu": {"": {"border": "1px solid black"}}}}',
        ...     reset=True
        ... )
        >>> s.get('Menu')
        {u'': {u'border': u'1px solid black'}}

        :param data: A JSON-encode ``str`` containing the data to load.
        :param reset: If ``True``, call `Styles.reset()` before updating styles
            so that the instance will only contain the style information
            provided in *data*
        """
        if reset:
            self.reset()
        d = json.loads(data)
        if not isinstance(d, dict):
            raise TypeError(
                'JSON data must define %r; got a %r: %r' % (dict, type(d), d)
            )
        if 'styles' in d:
            for (namespace, block) in d['styles'].iteritems():
                for (selector, definition) in block.iteritems():
                    self.update(namespace, selector, definition)
        if 'vars' in d:
            self.__vars.update(d['vars'])

    def diff(self, id, klass, substitute=True):
        """
        Return difference between the instance and the class CSS.

        This method returns the difference between the CSS of a widget instance
        and the CSS of the same widget's class.  This allows great configuration
        flexibility without rendering unnecessary CSS: because the instance
        inherits from the class, only the difference from the class to the
        instance needs to be rendered (if there is any difference at all).

        If no instance CSS has been defined, there is no difference and this
        method will return and empty ``dict``:

        >>> s = Styles()
        >>> s.update('Menu', '', {'color': 'red'})
        >>> s.diff('screen1.menuset.menu1', 'Menu')
        {}

        Similarly, there is no difference if the instance has only defined
        properties whose values match those in the class:

        >>> s.update('screen1.menuset.menu1', '', {'color': 'red'})
        >>> s.diff('screen1.menuset.menu1', 'Menu')
        {}

        However, if the instance defines properties not present in the class or
        properties with different values than those in the class, these
        differences are returned:

        >>> s.update('screen1.menuset.menu1', '', {'border': '1px solid green'})
        >>> s.diff('screen1.menuset.menu1', 'Menu')
        {'': {'border': '1px solid green'}}
        >>> s.update('screen1.menuset.menu1', '', {'color': 'black'})
        >>> s.diff('screen1.menuset.menu1', 'Menu')
        {'': {'color': 'black', 'border': '1px solid green'}}

        By default this method will perform variable substitution.  As the
        difference is calculated *after* the substitution, a variable that
        equals a corresponding literal will produce no difference:

        >>> s.update('Menu', '', {'border': '%(my_border)s'})
        >>> s.update('screen1.menuset.menu1', '', {'color': '%(my_color)s'})
        >>> s.update_vars(my_border='1px solid green', my_color='red')
        >>> s.diff('screen1.menuset.menu1', 'Menu')
        {}

        But if you call this method with ``substitute=False``, the differences
        in the raw styles will be revealed:

        >>> s.diff('screen1.menuset.menu1', 'Menu', substitute=False)
        {'': {'color': '%(my_color)s', 'border': '1px solid green'}}

        Like `Styles.get()`, this method will raise a `StyleVarError` if a
        CSS property references an undefined style-variable:

        >>> s.update('screen1.menuset.menu1', '', {'font-size': '%(my_font_size)s'})
        >>> s.diff('screen1.menuset.menu1', 'Menu')
        Traceback (most recent call last):
          ...
        StyleVarError: undefined style var 'my_font_size'
            used in CSS prop 'font-size' = '%(my_font_size)s'
            in namespace 'screen1.menuset.menu1', selector ''

        :param id: The id of widget instance (e.g., ``'screen1.menuset.menu1'``)
        :param klass: The widget's class name (e.g., ``'Menu'``)
        :param substitute: If ``True``, perform variable substitution
        """
        a = self.get(id, substitute)
        b = self.get(klass, substitute)
        if not a:
            {}
        if not b:
            return a
        return dict(
            dict_diff(a, b, depth=1)
        )

    def render(self, fileobj, tag, klass, id=None):
        """
        Render CSS for a given namespace.

        Without the optional ``id`` kwarg, the style is computed by calling
        `Styles.get()` and the class CSS is rendered.  For example:

        >>> s = Styles()
        >>> s.update('Menu', '', {'color': 'blue', 'background': '#fff'})
        >>> s.update('Menu', '> ul', {'position': 'fixed'})
        >>> s.render(sys.stdout, 'li', 'Menu')
        /*  ----  Menu  ----  */
        li.Menu {
            background: #fff;
            color: blue;
        }
        li.Menu > ul {
            position: fixed;
        }

        Which you can compare to the output of `Styles.get()`:

        >>> s.get('Menu')
        {'': {'color': 'blue', 'background': '#fff'}, '> ul': {'position': 'fixed'}}


        On the other hand, if you provide the ``id`` kwarg, the style is
        computed by calling `Styles.diff()` and the difference between the
        instance CSS and the class CSS is rendered.  For example:

        >>> s.update('screen1.menuset.menu1', '', {'color': 'red'})
        >>> s.update('screen1.menuset.menu1', '> ul', {'border': '1px solid green'})
        >>> s.render(sys.stdout, 'li', 'Menu', id='screen1.menuset.menu1')
        /*  ----  screen1.menuset.menu1  ----  */
        li#screen1.menuset.menu1 {
            color: red;
        }
        li#screen1.menuset.menu1 > ul {
            border: 1px solid green;
        }

        Which you can compare to the output of `Styles.diff()`:

        >>> s.diff('screen1.menuset.menu1', 'Menu')
        {'': {'color': 'red'}, '> ul': {'border': '1px solid green'}}


        If no style is defined for the namespace in question (of if the style is
        empty), nothing is written to ``fileobj``.  For example:

        >>> s.render(sys.stdout, 'div', 'TopBar')

        :param fileobj: An open file descriptor, ``StringIO`` instance, or similar
        :param tag: The widget's top-level HTML tag (e.g., ``'li'``)
        :param klass: The widget's class name (e.g., ``'Menu'``)
        :param id: The optional id of widget instance
                   (e.g., ``'screen1.menuset.menu1'``)
        """
        if id:
            unfiltered = self.diff(id, klass)
            namespace = id
            sep = '#'
        else:
            unfiltered = self.get(klass)
            namespace = klass
            sep = '.'
        prefix = sep.join([tag, namespace])
        css = tuple(filter_by_value(unfiltered, depth=1))
        if not css:
            return
        self._render_annotation(fileobj, namespace)
        for (selector, definition) in css:
            if selector:
                if selector[0] in (':.'):
                    j = ''
                else:
                    j = ' '
                full_selector = j.join([prefix, selector])
            else:
                full_selector = prefix
            self._render_rule(fileobj, full_selector, definition)

    def render_global(self, fileobj):
        """
        Render the global CSS (aka the ``''`` namespace).

        For example:

        >>> s = Styles()
        >>> s.update('', 'body', {'font-size': '11pt'})
        >>> s.update('', 'label.associate', {
        ...     'padding': '0.1em 0.2em',
        ...     'display': 'inline-block',
        ...     '-moz-appearance': 'button',
        ... })
        >>> s.update('', 'a', {'text-decoration': 'none'})
        >>> s.render_global(sys.stdout)
        /*  ----  [GLOBAL]  ----  */
        a {
            text-decoration: none;
        }
        body {
            font-size: 11pt;
        }
        label.associate {
            -moz-appearance: button;
            display: inline-block;
            padding: 0.1em 0.2em;
        }

        :param fileobj: An open file descriptor, ``StringIO`` instance, or similar
        """
        css = tuple(filter_by_value(self.get(''), depth=1))
        if not css:
            return
        self._render_annotation(fileobj, '[GLOBAL]')
        for (full_selector, definition) in css:
            self._render_rule(fileobj, full_selector, definition)

    def _substitute(self, namespace, selector):
        """
        Substitute variables.
        """
        if namespace not in self.__tree:
            return
        ns = self.__tree[namespace]
        if selector not in ns:
            return
        definition = ns[selector]
        for (key, value) in definition.iteritems():
            if not value:
                yield (key, value)
            else:
                try:
                    value %= self.__vars
                    yield (key, value)
                except KeyError, e:
                    raise StyleVarError(
                        var=e.args[0],
                        key=key,
                        value=value,
                        namespace=namespace,
                        selector=selector,
                    )

    def _render_annotation(self, fileobj, namespace):
        """
        Write a standard CSS comment to indicate the start a new style namespace.

        >>> s = Styles()
        >>> s._render_annotation(sys.stdout, 'Menu')
        /*  ----  Menu  ----  */
        >>> s._render_annotation(sys.stdout, 'screen1.menuset.menu1')
        /*  ----  screen1.menuset.menu1  ----  */

        :param fileobj: An open file descriptor, ``StringIO`` instance, or similar
        :param namespace: The widget class name or instance id (e.g., ``'Menu'``
                          or ``'screen1.menuset.menu1'``)
        """
        fileobj.write('/*  ----  %s  ----  */\n' % namespace)

    def _render_rule(self, fileobj, full_selector, definition):
        """
        Render a CSS rule.

        For example:

        >>> s = Styles()
        >>> definition = (
        ...     ('display', 'none'),
        ...     ('position', 'fixed'),
        ... )
        >>> s._render_rule(sys.stdout, 'li.Menu:hover > ul', definition)
        li.Menu:hover > ul {
            display: none;
            position: fixed;
        }

        :param fileobj: An open file descriptor, ``StringIO`` instance, or similar
        :param full_selector: The CSS selector (e.g., ``'li.Menu:hover > ul'``)
        :param definition: Iterable of ``(key,value)`` tuples
        """
        fileobj.write('%s {\n' % full_selector)
        for (key, value) in definition:
            fileobj.write('    %s: %s;\n' % (key, value))
        fileobj.write('}\n')
